#!/usr/bin/python
# Copyright 2016 CERN and GSI
#
# This program is free software: you can redistribute it and/or modify
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import time
import zlib
import libxml2

import iecommon
import iefiles
import xmltemplate
import s7template
import virtualS7Template
import rabbitTemplate

from iecommon import *
from model.Class.Register import ParamRegister
from model.Class.Block import ParamBlock
from model.Class.Class import ParamClass
from model.Deploy.Deploy import *

#=========================================================================
# General remarks
#=========================================================================

#XML input documents well-formed is not checked! 
#This Python-script generates PLC sources document from XML Parameter documents
#which are supposed to be well-formed (generated with the 'genparam' script).


# Deploy global information
class DeploymentInfo(object):
    plcModel        = None  # S7-300, .., Premium, ..
    plcSystem       = None  # UNITY, STEP-7, TIA-PORTAL, ...
    plcProtocol     = None  # DEVICE_MODE or BLOCK_MODE
    address         = None
    checksum     = None  # checksum of the related parameter file
    owner           = None  # onwer of the Deployment document
    silecsVersion   = None

whichUnityFormat = {
    'uint8'      : 'BYTE',
    'int8'        : 'WORD',
    'uint16'    : 'WORD',
    'int16'      : 'INT',
    'uint32'    : 'DWORD',
    'int32'      : 'DINT',
    'float32'   : 'REAL',
    'string'     : 'STRING',
#    'uint64'      : '8',    not supported with PLC
#    'int64'        : '8',    not supported with PLC
#    'float64'     : '8',    not supported with PLC
    'date'       : 'DT',
    'char'       : 'WORD',
    'byte'       : 'BYTE',
    'word'      : 'WORD',
    'dword'    : 'DWORD',
    'int'          : 'INT',
    'dint'        : 'DINT',
    'real'        : 'REAL',
    'dt'           : 'DT'
}

whichTwincatFormat = {
    'uint8'      : 'BYTE',
    'int8'        : 'SINT',
    'uint16'    : 'WORD',
    'int16'      : 'INT',
    'uint32'    : 'DWORD',
    'int32'      : 'DINT',
    'float32'   : 'REAL',
    'string'     : 'STRING',
#    'uint64' not supported by PLCs
#    'int64' not supported by PLCs
#    'float64' not supported by PLCs
    'date'       : 'DT',
    'char'       : 'SINT',
    'byte'       : 'BYTE',
    'word'      : 'WORD',
    'dword'    : 'DWORD',
    'int'          : 'INT',
    'dint'        : 'DINT',
    'real'        : 'REAL',
    'dt'           : 'DT',
    'string'     : 'STRING'
}

schneiderRegArray  = """ARRAY[0..%d] OF %s"""
schneiderRegArray2d  = """ARRAY[0..%d, 0..%d] OF %s"""

# Beckhoff templates (using newlines \r\n because TwinCAT compiler complains with line feed character)
beckhoffReg = """    %s_%04x_%s AT %%MW%s: %s;\r\n\r\n"""
beckhoffRegArray = """    %s_%04x_%s AT %%MW%s: ARRAY [0..%d] OF %s;\r\n\r\n"""
beckhoffRegArray2d = """    %s_%04x_%s AT %%MW%s: ARRAY [0..%d, 0..%d] OF %s;\r\n\r\n"""
beckhoffRegInit = """    %s_%04x_%s AT %%MW%s: %s:= %s;\r\n\r\n"""

#=========================================================================
# Sub-function
#=========================================================================

#-------------------------------------------------------------------------
# Schneider registers
#-------------------------------------------------------------------------

def xsyRegister(register):
    if register.isStringType():
        strLen = whichUnityFormat[register.format]+'['+str(register.stringLength)+']'
        if  register.isString():
            return strLen
        elif register.isStringArray():
            return schneiderRegArray %(register.dim1-1, strLen)
        else: #2D-Array
            return schneiderRegArray2d %(register.dim1-1, register.dim2-1, strLen)
    else:
        if  register.isScalar():
            return whichUnityFormat[register.format]
        elif register.isArray():
            return schneiderRegArray %(register.dim1-1, whichUnityFormat[register.format])
        else: #2D-Array
            return schneiderRegArray2d %(register.dim1-1, register.dim2-1, whichUnityFormat[register.format])
    
def getDateTime():
    dt = time.localtime(time.time())
    return "DT#%s-%s-%s-%s:%s:%s" %(dt[0],dt[1],dt[2],dt[3],dt[4],dt[5])
            
#-------------------------------------------------------------------------
# SIEMENS PLC code generation
#-------------------------------------------------------------------------
def generateSiemensSources(paramDOM, sourceFolderPath ,logTopics):
    stlString = ''
    symString = ''  # for .sdf Symbol file

    deploy = getDeploymentInformation(paramDOM)

    # Prepare the Simatic Symbols (<.sdf> file)
    DBnumber    = long(deploy.address)       #first data-block (SilecsHeader) starts from the base-address
    UDTnumber   = long(deploy.address)       #use base data-block address for UDT number as well (to be reserved by the developer)

    # Generate sources for each class that is deployed in the PLC
    for paramClass in ParamClass.getParamClassesFromRootNode(paramDOM):
        deviceDOMList = paramClass.getDeviceInstanceNodes() # get Device instances of that class
        deviceNumber = len(deviceDOMList)

        # Generate the Blocks definition
        for blockIndex, block in enumerate(paramClass.getParamBlocks()):
            registerList = ''
            for register in block.getParamRegisters():
                # If PLC does not supports this register format abort this generation and log the error
                if(register.format in ['uint64','int64','float64']):
                    iecommon.logError('ERROR: In design %s_%s register %s, %s format not supported for current controller model.'
                                      %(paramClass.name, paramClass.version, register.name, register.format)
                                      , True,logTopics)
                
                #create register value if required (diagnostic registers assignment in particular)
                registerValue = s7template.stlRegisterValue(register.name, deploy.owner, getDateTime(), deploy.checksum,deploy.silecsVersion)
                
                if register.format == 'string':
                    registerList += s7template.stlRegister(register.name, register.format, register.dim1, register.dim2, registerValue, long(register.stringLength))
                else:                    
                    registerList += s7template.stlRegister(register.name, register.format, register.dim1, register.dim2, registerValue)

            stlString += s7template.stlBlockUDT(deploy.owner, paramClass.name, paramClass.version, block.name, registerList, (blockIndex == 0))
            symString += s7template.symBlockUDT(paramClass.name, paramClass.version, block.name, UDTnumber, (blockIndex == 0))
            UDTnumber += 1
   
        if deploy.plcProtocol == 'DEVICE_MODE':
        # Generate the Data-Blocks: one DB per Device instance
            for deviceIndex, deviceDOM in enumerate(deviceDOMList):
                blockList = ''
                for block in paramClass.getParamBlocks():
                    blockList += s7template.stlBlock(paramClass.name, block.name)
                deviceLabel = deviceDOM.prop('label')
                stlString += s7template.generateBlock(deploy.owner, paramClass.name, paramClass.version, deploy.plcSystem, DBnumber, deviceLabel, blockList, (deviceIndex == 0),deploy.plcProtocol)
                symString += s7template.symDeviceDB(paramClass.name, paramClass.version, deviceLabel, DBnumber, (deviceIndex == 0))
                DBnumber += 1

        else:             # BLOCK_MODE
        # Generate the Data-Blocks: one DB per Block of registers
            for blockIndex, block in enumerate(paramClass.getParamBlocks()):
                deviceList = ''
                for deviceDOM in deviceDOMList:
                    deviceLabel = deviceDOM.prop('label')
                    deviceList += s7template.stlDevice(deviceLabel, paramClass.name, block.name)

                stlString += s7template.generateBlock(deploy.owner, paramClass.name, paramClass.version, deploy.plcSystem, DBnumber, block.name, deviceList, (blockIndex == 0),deploy.plcProtocol)
                symString += s7template.symBlockDB(paramClass.name, paramClass.version, block.name, DBnumber, (blockIndex == 0))
                DBnumber += 1

    # write the String PLC generated code
    generateControllerFiles(sourceFolderPath,deploy,".scl",stlString,logTopics);
        
    # Write the String PLC generated symbols into the <.sdf> symbols file
    generateControllerFiles(sourceFolderPath,deploy,".sdf",symString,logTopics);

#-------------------------------------------------------------------------
# SCHNEIDER PLC code generation
#-------------------------------------------------------------------------

# ----------------------------------------------------
def generateSchneiderRegisters(xsydoc, paramClass, deviceDOM, block, deviceIndex, modeDevice, deploy, logTopics):
    
    deviceLabel = deviceDOM.prop('label')
    
    #Device and Block adresses come from the generated Paramerers file. Depends on the protocol mode:
    #DEVICE_MODE: relative address of the element within it parent node
    #BLOCK_MODE : absolute address in the entire PLC memory
    deviceAddress = long(deviceDOM.prop('address'))

    for register in block.getParamRegisters():
        # If PLC does not supports this register format abort this generation and log the error
        if(register.format in ['uint64','int64','float64']):
            iecommon.logError('ERROR: In register '+register.name+', '+register.format+' format not supported for current controller model.', True,logTopics)
        
        # Compute the Register address relying on the Class Parameters file data
        # Attention! Schneider uses WORD addressing while Parameters data are expressed in bytes to be PLC independent
        if modeDevice:
            totalAddress = (deviceAddress + block.address + register.address)/2
        else:
            totalAddress = (block.address + (deviceIndex * block.memSize) + register.address)/2
              
        # then insert the corresponding DOM element
        dataElt = xsydoc.xpathEval("//dataBlock")[0] #only one dataBlock node has been inserted so far
        regElt  = libxml2.newNode('variables')
        # Compute the base checksum of the class name
        classNameCRC = zlib.crc32(paramClass.name,0) & 0xffff
        # Create register name = regname_crc32(classname)_deviceid
        # 24 char (register name) + 1 (underscore) + 4 char (CRC classname) + 1 (underscore) + n (device id) = 30 + n (device id) 
        # possible problem when device id (n)> 99 --> string > 32 not compatible with Schneider PLCs
        regElt.setProp("name", "%s_%04x_%s" %(register.name, classNameCRC, deviceLabel))
        
        if register.format == 'string':
            regElt.setProp("typeName", "%s" %xsyRegister(register))
        else:
            regElt.setProp("typeName", "%s" %xsyRegister(register))
        
            
        regElt.setProp("topologicalAddress", "%%MW%ld" %totalAddress)
        if block.name == 'hdrBlk':
            initElt = libxml2.newNode('variableInit')
            if register.name == '_version': initElt.setProp("value", deploy.silecsVersion )
            if register.name == '_user'   : initElt.setProp("value", deploy.owner)
            if register.name == '_checksum': initElt.setProp("value", "%s" %deploy.checksum)
            if register.name == '_date'  : initElt.setProp("value", getDateTime())
            regElt.addChild(initElt)
        commElt    = libxml2.newNode('comment')
        commElt.setContent(paramClass.name+"/"+deviceLabel+"/"+block.name)
        regElt.addChild(commElt)
        dataElt.addChild(regElt)
        
# ----------------------------------------------------
def generateSchneiderSources(paramDOM,sourceFolderPath,logTopics):
    #Base address provided by the user in the Deployment relies on the PLC model:  
    #Quantum, Premium and M340 Schneider PLCs are all interpreted as using 16-bit addressing, 
    #even though M340 addresses are later properly considered as 32-bit addresses
    
    deploy = getDeploymentInformation(paramDOM)

    # Create the Schneider DOM source (Unity <.XSY> XML document)
    xsyDoc  = libxml2.newDoc(version='1.0')
    rootElt = libxml2.newNode('VariablesExchangeFile')
    xsyDoc.addChild(rootElt)
    dataElt = libxml2.newNode('dataBlock')
    rootElt.addChild(dataElt)

    for paramClass in ParamClass.getParamClassesFromRootNode(paramDOM):
        deviceDOMList = paramClass.getDeviceInstanceNodes() # get Device instances of that class
        
        for block in paramClass.getParamBlocks():
            for deviceIndex,deviceDOM in enumerate(deviceDOMList):
                if deploy.plcProtocol == 'DEVICE_MODE': #-------------------
                    generateSchneiderRegisters(xsyDoc, paramClass, deviceDOM, block, deviceIndex, True, deploy, logTopics)
                else:            
                    generateSchneiderRegisters(xsyDoc, paramClass, deviceDOM, block, deviceIndex, False, deploy, logTopics)
    
    # Finally, write the DOM object (PLC generated code) into the XSY source file
    generateControllerFiles(sourceFolderPath,deploy,".xsy",rootElt.serialize(format = True),logTopics);

    xsyDoc.freeDoc()      
    
# ----------------------------------------------------
def generateRabbitSources(paramDOM,sourceFolderPath,logTopics):   
    deviceLabel = ''

    deploy = getDeploymentInformation(paramDOM)

    # Prepare the source (<.h> file) with its diagnostic data-block header
    cCodeString =  rabbitTemplate.cHeader(deploy.address,deploy.silecsVersion)
    cTypeDefinitions = ''
    cDataAllocation  = rabbitTemplate.cAllocationComment(deploy.plcProtocol)
    
    blockList = ''
    classList = ''
    NBdeviceDefinitionList = ''

    for paramClass in ParamClass.getParamClassesFromRootNode(paramDOM):
        deviceDOMList = paramClass.getDeviceInstanceNodes() # get Device instances of that class
        deviceNumber = len(deviceDOMList)

        #==== Generate the Blocks definition ====
        for blockIndex, block in enumerate(paramClass.getParamBlocks()):
            registerList = ''
            for register in block.getParamRegisters():
                # If PLC does not supports this register format abort this generation and log the error
                if(register.format in ['float64']):
                    iecommon.logError('ERROR: In register '+ register.name +', '+register.format+' format not supported for current controller model.', True,logTopics)
                
                if(register.format == 'string'):
                    registerList += rabbitTemplate.cRegister(register.name, register.format, register.dim1, register.dim2, long(register.stringLength))
                else:
                    registerList += rabbitTemplate.cRegister(register.name, register.format, register.dim1, register.dim2)    

            cTypeDefinitions += rabbitTemplate.cBlockUDT(paramClass.name, paramClass.version, block.name, registerList, (blockIndex == 0))
            
        #==== Memory allocation ====
        if deploy.plcProtocol == 'DEVICE_MODE':
            # Generate the List of block in the class
            blockList =''
            for block in paramClass.getParamBlocks():
                blockList += rabbitTemplate.cDeviceModeBlockInstantiation(paramClass.name, block.name)

            deviceList = ''
            for deviceDOM in deviceDOMList:
                deviceLabel = paramClass.name + "_" + deviceDOM.prop('label')
                deviceList += deviceLabel + ', ';
            deviceList = deviceList[:-2]  # remove last comma space
            classList += rabbitTemplate.cDeviceModeClass_deviceList(blockList, deviceList)


        else:  # BLOCK_MODE
            # Generate the List of block in the class
            for block in paramClass.getParamBlocks():
                deviceList = ''
                for deviceDOM in deviceDOMList:
                    deviceLabel = deviceDOM.prop('label')
                    deviceList +=rabbitTemplate.cBlockModeDeviceInstantiation_deviceList(paramClass.name, block.name, deviceLabel)

                blockList += rabbitTemplate.cBlockModeBlockInstantiation(deviceList, paramClass.name, block.name)

    # Predefine number of devices constant
    cDataAllocation += NBdeviceDefinitionList                

    if deploy.plcProtocol == 'DEVICE_MODE':
        cDataAllocation += rabbitTemplate.cDeviceModeDataInstantiation(classList)
        cInitFunction    = rabbitTemplate.cInitDeviceMode(deploy.silecsVersion,deploy.checksum,deploy.owner)
    else:
        cDataAllocation += rabbitTemplate.cBlockModeDataInstantiation(blockList)
        cInitFunction    = rabbitTemplate.cInitBlockMode(deploy.silecsVersion,deploy.checksum,deploy.owner)
    
            
    # Append data definitions
    cCodeString += cTypeDefinitions
    
    # Append data allocation
    cCodeString += cDataAllocation
    
    # Append initialization function
    cCodeString += cInitFunction
    
    # Generate and append sample example
    cCodeString += rabbitTemplate.cExample(deploy.plcProtocol,True,paramClass.name, block.name, register.name, deviceLabel)
    
    # Finally, write the String C generated code into the temporal output file
    generateControllerFiles(sourceFolderPath,deploy,".h",cCodeString,logTopics);


# ----------------------------------------------------
def generatePCLikeSources(paramDOM,sourceFolderPath,logTopics):   
    deviceLabel = ''

    deploy = getDeploymentInformation(paramDOM)

    # Prepare the source (<.h> file) with its diagnostic data-block header
    cCodeString =  PCLikeTemplate.cHeader(deploy.baseAddress)
    cTypeDefinitions = ''
    cDataAllocation  = PCLikeTemplate.cAllocationComment(deploy.plcProtocol)
    
    blockList = ''
    classList = ''
    NBdeviceDefinitionList = ''

    # Generate sources for each class that is deployed in the PLC
    for paramClass in ParamClass.getParamClassesFromRootNode(paramDOM):

        deviceNumber = len(deviceDOMList)

        #==== Generate the Blocks definition ====
        for blockIndex, blockDOM in enumerate(blockDOMList):
            registerList = ''
            blockName = blockDOM.getAttribute('name')
            registerDOMList = blockDOM.getElementsByTagName("Register") # get Registers of the current block
            for registerIndex, registerDOM in enumerate(registerDOMList):
                registerName = registerDOM.getAttribute('name')
                registerFormat = registerDOM.getAttribute('format')
                registerDimension1 = long(registerDOM.getAttribute('array-dim1'))
                registerDimension2 = long(registerDOM.getAttribute('array-dim2'))
                
                # If PLC does not supports this register format abort this generation and log the error
                if(registerFormat in ['float64']):
                    iecommon.logError('ERROR: In register '+registerName+', '+registerFormat+' format not supported for current controller model.', True,logTopics)
                
                if(registerFormat == 'string'):
                    registerList += PCLikeTemplate.cRegister(registerName, registerFormat, registerDimension1, registerDimension2, long(registerDOM.getAttribute('string-len')))
                else:
                    registerList += PCLikeTemplate.cRegister(registerName, registerFormat, registerDimension1, registerDimension2)    

            cTypeDefinitions += PCLikeTemplate.cBlockUDT(className, classVersion, blockName, registerList, (blockIndex == 0))
            
        #==== Memory allocation ====
        if deploy.plcProtocol == 'DEVICE_MODE':
            # Generate the List of block in the class
            blockList =''
            for blockIndex, blockDOM in enumerate(blockDOMList):
                blockName = blockDOM.getAttribute('name')
                blockList += PCLikeTemplate.cDeviceModeBlockInstantiation(className, blockName)
            # instantiate for deviceList / deviceNumber
            if withDeviceList == True:
                # DeviceList
                deviceList = ''
                for deviceDOM in deviceDOMList:
                    deviceLabel = className + "_" + deviceDOM.getAttribute('label')
                    deviceList += deviceLabel + ', ';
                deviceList = deviceList[:-2]  # remove last comma space
                classList += PCLikeTemplate.cDeviceModeClass_deviceList(blockList, deviceList)
            else :
                # DeviceNumber
                classList += PCLikeTemplate.cDeviceModeClass_deviceNumber(className, blockList)
                NBdeviceDefinitionList += PCLikeTemplate.cNBdeviceDefinition(className,deviceNumber)

        else:  # BLOCK_MODE
            # Generate the List of block in the class
            for blockIndex, blockDOM in enumerate(blockDOMList):
                # create instantation for current block
                blockName = blockDOM.getAttribute('name')
                if withDeviceList == True:
                    # DeviceList
                    deviceList = ''
                    for deviceDOM in deviceDOMList:
                        deviceLabel = deviceDOM.getAttribute('label')
                        deviceList +=PCLikeTemplate.cBlockModeDeviceInstantiation_deviceList(className, blockName, deviceLabel)
                else:
                    # DeviceNumber
                    deviceList = PCLikeTemplate.cBlockModeDeviceInstantiation_deviceNumber(className, blockName)

                # allocation is the same with device list or device number
                blockList += PCLikeTemplate.cBlockModeBlockInstantiation(deviceList, className, blockName)
            
            if withDeviceList == False:
                NBdeviceDefinitionList += PCLikeTemplate.cNBdeviceDefinition(className,deviceNumber)

    # Predefine number of devices constant
    cDataAllocation += NBdeviceDefinitionList                

    if deploy.plcProtocol == 'DEVICE_MODE':
        cDataAllocation += PCLikeTemplate.cDeviceModeDataInstantiation(classList)
        cInitFunction    = PCLikeTemplate.cInitDeviceMode(ieglobal.version,deploy.checksumRef,deploy.owner)
    else:
        cDataAllocation += PCLikeTemplate.cBlockModeDataInstantiation(blockList)
        cInitFunction    = PCLikeTemplate.cInitBlockMode(ieglobal.version,deploy.checksumRef,deploy.owner)
    
              
    # Append data definitions
    cCodeString += cTypeDefinitions
    
    # Append data allocation
    cCodeString += cDataAllocation
    
    # Append initialization function
    cCodeString += cInitFunction
    
    # Generate and append sample example
    cCodeString += PCLikeTemplate.cExample(deploy.plcProtocol,withDeviceList,className, blockName, registerName, deviceLabel)
    
    # Finally, write the String C generated code into the temporal output file
    sourcesPath = iefiles.getControllerSourcesDirectory(workspacePath, deploy.plcName, deploy.deployVersion)
    sourcesFile = os.path.normpath(sourcesPath + "/%s.h" % deploy.plcName)
    iecommon.logInfo("Generate PLC sources: %s" %sourcesFile,logTopics);
    fdesc = open(sourcesFile, "w")
    fdesc.write(cCodeString)
    fdesc.close()
    
    # Clean-up the memory before return
    deployDOM.unlink()

# ----------------------------------------------------
def generateVirtualS7Sources(deployDOM, classDOCList, deploy, workspacePath,logTopics):   

    includeDesign  = ''
    allocDesign  = ''
    deleteDesign = ''
    getDesign = ''

    protocolMode = 'BlockMode'; 
    if deploy.plcProtocol == 'DEVICE_MODE':
        protocolMode = 'DeviceMode'; 
    duConstructor = virtualS7Template.vs7DuConstructor(deploy.plcName, deploy.deployVersion, protocolMode, deploy.baseAddress)
    
    # Generate sources for each class that is deployed in the PLC
    for classDOC in classDOCList:
        blocksCodeString = ''
         
        createBlockCodeString = '' 
        deleteBlockCodeString = '' 
        deviceCodeString = ''
         
        createDeviceCodeString = '' 
        deleteDeviceCodeString = '' 
        designCodeString = ''
         
        classDOM = classDOC.getElementsByTagName("SILECS-Class")[0]  #only one SILECS-Class node is expected
        blockDOMList  = classDOC.getElementsByTagName("Block")       # get Blocks elements of that class
        
        className       = classDOM.getAttribute('name')
        classVersion    = classDOM.getAttribute('version')
    
        #DEPLOY-UNIT code generation ============================
        includeDesign += virtualS7Template.vs7DuDesignInclude(className, classVersion)   
        allocDesign += virtualS7Template.vs7DuDesignAlloc(className, classVersion)   
        deleteDesign += virtualS7Template.vs7DuDesignDelete(className)
        getDesign += virtualS7Template.vs7DuDesignGet(className, classVersion)

        #CLASSES code generation ================================
        for blockIndex, blockDOM in enumerate(blockDOMList):
            getSetCodeString = ''
            dimsCodeString   = ''
            dataCodeString   = ''
            
            currentRegisterAddress = 0  #used to compute even/odd adress and insert align data by adding dummy registers
            previousRegisterMemSize = 0 #...
            dummyIndex = 0;
             
            blockName = blockDOM.getAttribute('name')
            blockOffset = blockDOM.getAttribute('address')
            registerDOMList = blockDOM.getElementsByTagName("Register") # get Registers of the current block
            
            createBlockCodeString += virtualS7Template.vs7ClassCreateBlock(className, blockName)
            deleteBlockCodeString += virtualS7Template.vs7ClassDeleteBlock(className, blockName)
            
            for registerIndex, registerDOM in enumerate(registerDOMList):
                registerName = registerDOM.getAttribute('name')
                registerFormat = registerDOM.getAttribute('format')
                registerDimension1 = long(registerDOM.getAttribute('array-dim1'))
                registerDimension2 = long(registerDOM.getAttribute('array-dim2'))
                registerMemSize = long(registerDOM.getAttribute('mem-size')) #store the reigster mem-size for next iteration
                
                registerLength = 1  #in case string 
                if(registerFormat == 'string'):
                    registerLength = long(registerDOM.getAttribute('string-len'))
            
                getSetCodeString += virtualS7Template.vs7ClassGetSet(registerName, registerFormat, registerDimension1, registerDimension2,registerLength)
                dimsCodeString += virtualS7Template.vs7ClassDimension(registerName, registerFormat, registerDimension1, registerDimension2,registerLength)

                #Specific code to force 16bit alignment if not scalar 8bit register (==> insert dummy register if needed)
                dummyCodeString = virtualS7Template.vs7ClassDummyRegister(currentRegisterAddress, previousRegisterMemSize, registerMemSize, dummyIndex)
                dataCodeString += dummyCodeString
                currentRegisterAddress += registerMemSize 
                if dummyCodeString != "":
                    dummyIndex += 1
                    currentRegisterAddress += 1     
                #and now insert the normal data register     
                dataCodeString += virtualS7Template.vs7ClassDataRegister(registerName, registerFormat, registerDimension1, registerDimension2, registerLength)
                previousRegisterMemSize = registerMemSize 
                
            blocksCodeString += virtualS7Template.vs7ClassBlock(ieglobal.version, deploy.owner, deploy.checksumRef, className, blockName, getSetCodeString, blockOffset, dimsCodeString, dataCodeString)
        
        deviceDOMList = classDOC.getElementsByTagName("Instance") # get Device instances of that class
        for deviceIndex, deviceDOM in enumerate(deviceDOMList):
            deviceLabel = deviceDOM.getAttribute('label')
            createDeviceCodeString += virtualS7Template.vs7ClassCreateDevice(className, deviceLabel, deviceIndex)
            deleteDeviceCodeString += virtualS7Template.vs7ClassDeleteDevice(className, deviceLabel)
        
        deviceCodeString = virtualS7Template.vs7ClassDevice(createBlockCodeString, deleteBlockCodeString)
        designCodeString = virtualS7Template.vs7ClassDesign(className, classVersion, createDeviceCodeString, deleteDeviceCodeString)
        
        classCodeString = virtualS7Template.vs7ClassHeader(className, classVersion, blocksCodeString, deviceCodeString, designCodeString)
            
        # Finally, write the CLASS generated code into the temporal output file
        sourcesPath = iefiles.getControllerSourcesDirectory(workspacePath, deploy.plcName, deploy.deployVersion)
        sourcesFile = os.path.normpath(sourcesPath + "/%s_%s.h" % (className, classVersion))
        iecommon.logInfo("Generate PLC sources: %s" %sourcesFile,logTopics);
        fdesc = open(sourcesFile, "w")
        fdesc.write(classCodeString)
        fdesc.close()
    
    # Prepare the source (<.h> file) with its diagnostic data-block header
    duCodeString = virtualS7Template.vs7DuHeader(deploy.plcName, deploy.deployVersion, duConstructor, includeDesign, allocDesign, deleteDesign, getDesign)    
             
    # Finally, write the DEPLOY-UNIT generated code into the temporal output file
    sourcesPath = iefiles.getControllerSourcesDirectory(workspacePath, deploy.plcName, deploy.deployVersion)
    sourcesFile = os.path.normpath(sourcesPath + "/%s_%s.h" % (deploy.plcName, deploy.deployVersion))
    iecommon.logInfo("Generate PLC sources: %s" %sourcesFile,logTopics);
    fdesc = open(sourcesFile, "w")
    fdesc.write(duCodeString)
    fdesc.close()
             
    # Clean-up the memory before return
    deployDOM.unlink()
    
# ----------------------------------------------------
def generateBeckhoffRegisters(paramClass, deviceDOM, block, deviceIndex, deploy, logTopics):
    source = ''    
    deviceLabel = deviceDOM.prop('label')
    
    #Device and Block adresses come from the generated Paramerers file. Depends on the protocol mode:
    #DEVICE_MODE: relative address of the element within it parent node
    #BLOCK_MODE : absolute address in the entire PLC memory
    deviceAddress = long(deviceDOM.prop('address'))
    
    for register in block.getParamRegisters():
        # If PLC does not supports this register format abort this generation and log the error
        if(register.format in ['uint64','int64','float64']):
            iecommon.logError('ERROR: In register '+register.name+', '+register.format+' format not supported for current controller model.', True)
            return
    
        iecommon.logDebug("Processing: %s %s %s %s" %(paramClass.name, deviceLabel, block.name, register.name))

        # Compute the register address relying on the Class Parameters file data
        # Attention! Beckhoff uses WORD addressing while Parameters data are expressed in bytes to be PLC independent
        if deploy.plcModel == 'BC9020':
            totalAddress = (block.address - long(deploy.address) + (deviceIndex * block.memSize) + register.address)/2
        elif deploy.plcModel == 'CX9020':
            totalAddress = (block.address - long(deploy.address) + (deviceIndex * block.memSize) + register.address)
        else:
            raise "PLC model not supported: " + deploy.plcModel

        # add comment to source to identify block
        source += '    (*'+paramClass.name+'/'+deviceLabel+'/'+block.name+' *)\r\n'
        
        # Compute the base checksum of the class name
        classNameCRC = zlib.crc32(paramClass.name,0) & 0xffff
        
        if block.name == 'hdrBlk':
            if register.name == '_version':
                source+=beckhoffRegInit %(register.name, classNameCRC, deviceLabel, totalAddress, whichTwincatFormat[register.format]+"(16)", deploy.silecsVersion )
            if register.name == '_user':
                source+=beckhoffRegInit %(register.name, classNameCRC, deviceLabel, totalAddress, whichTwincatFormat[register.format]+"(16)", "'"+deploy.owner+"'") 
            if register.name == '_checksum':
                source+=beckhoffRegInit %(register.name, classNameCRC, deviceLabel, totalAddress, whichTwincatFormat[register.format], deploy.checksum)
            if register.name == '_date':
                source+=beckhoffRegInit %(register.name, classNameCRC, deviceLabel, totalAddress, whichTwincatFormat[register.format], getDateTime())
        
        else:   # not header block
             # set data type - for string with fixed value registerLength
            if register.format == 'string':
                format = whichTwincatFormat[register.format]+'('+str(register.stringLength)+')'
            else:
                format = whichTwincatFormat[register.format]
                
            if register.dim1 == 1 and register.dim2 == 1:    # scalar
                source += beckhoffReg %(register.name, classNameCRC, deviceLabel, totalAddress, format)
            elif register.dim1 > 1 and register.dim2 == 1: # array
                source += beckhoffRegArray %(register.name, classNameCRC, deviceLabel, totalAddress, register.dim1-1, format)
            else:   # dim1>=1 and for whatever dim2, use double array syntax
                source += beckhoffRegArray2d %(register.name, classNameCRC, deviceLabel, totalAddress, register.dim1-1, register.dim2-1, format)
                    
    return source
# ----------------------------------------------------
def generateBeckhoffSources(paramDOM, sourceFolderPath, logTopics):
    deploy = getDeploymentInformation(paramDOM)
    
    # Prepare the global variables source (<.exp> file) with its diagnostic data-block header
    source ="""(* +-------------------------------------------------------------------\r\n"""
    source += """* | C.E.R.N Geneva, Switzerland\r\n"""
    source += """* | SILECS - BE/CO-FE\r\n"""
    source += """* | April 2015\r\n"""
    source += """* +-------------------------------------------------------------------\r\n"""
    source += """*\r\n"""
    source += """* Release : SILECS_%s\r\n"""%(deploy.silecsVersion)
    source += """*)"""
        
    source += '\r\nVAR_GLOBAL\r\n'

    for paramClass in ParamClass.getParamClassesFromRootNode(paramDOM):
        deviceDOMList = paramClass.getDeviceInstanceNodes() # get Device instances of that class
            
        # Device mode is not supported for Beckhoff PLCs - Only block mode available
        for block in paramClass.getParamBlocks():
            for deviceIndex, deviceDOM in enumerate(deviceDOMList):
                source += generateBeckhoffRegisters(paramClass, deviceDOM, block, deviceIndex,deploy,logTopics)
                
    source += 'END_VAR'

    # Write the source into the EXP source file
    generateControllerFiles(sourceFolderPath,deploy,".exp",source,logTopics);

#-------------------------------------------------------------------------
# NI code generation
#-------------------------------------------------------------------------
def generateNISources(paramDOM, sourceFolderPath ,logTopics):
    iecommon.logInfo("PLC source generation not available for National Instruments equipment",logTopics)
    return

def getDeploymentInformation(paramDOM):
    deploy = DeploymentInfo()
    mapping = paramDOM.xpathEval("/SILECS-Param/SILECS-Mapping")[0] 

    deploy.plcModel    = mapping.prop('plc-model')
    deploy.plcSystem    = mapping.prop('plc-system')
    deploy.plcProtocol = mapping.prop('protocol')
    deploy.plcName = mapping.prop('plc-name')
    deploy.address = mapping.prop('address')
    ownerNode = paramDOM.xpathEval("/SILECS-Param/Mapping-Info/Owner")[0] 
    deploy.owner = ownerNode.prop('user-login')
    deploymentNode = paramDOM.xpathEval("/SILECS-Param/Mapping-Info/Deployment")[0] 
    deploy.checksum = long(deploymentNode.prop('checksum'))
    rootNode = paramDOM.xpathEval("/SILECS-Param")[0] 
    deploy.silecsVersion = rootNode.prop("silecs-version")
    return deploy

def generateControllerFiles(sourceFolderPath,deploy,fileExtention,source,logTopics={'errorlog': True}):
    fileName = deploy.plcName + fileExtention
    fullFilePath = os.path.normpath(sourceFolderPath + "/" + fileName )

    iecommon.logInfo("Generate PLC sources: %s" %fullFilePath,logTopics);
    fdesc = open(fullFilePath, "w")
    fdesc.write(source)
    fdesc.close()

# Needed for unit-testing
def generateControllerCode(controller, paramsFile, sourceFolderPath, logTopics={'errorlog': True}):
    paramDOM = libxml2.parseFile(paramsFile)
    if controller.plcNode.get_name() == 'Siemens-PLC':
        generateSiemensSources(paramDOM,sourceFolderPath,logTopics)
    elif controller.plcNode.get_name() == 'PC-Controller':
        generatePCLikeSources(paramDOM,sourceFolderPath,logTopics)
    elif controller.plcNode.get_name() == 'VirtualS7':
        generateVirtualS7Sources(paramDOM,sourceFolderPath,logTopics)
    elif controller.plcNode.get_name() == 'Schneider-PLC':
        generateSchneiderSources(paramDOM,sourceFolderPath,logTopics)
    elif controller.plcNode.get_name() == 'Rabbit-uC':
        generateRabbitSources(paramDOM,sourceFolderPath,logTopics)
    elif controller.plcNode.get_name() == 'Beckhoff-PLC':
        generateBeckhoffSources(paramDOM,sourceFolderPath,logTopics)
    elif controller.plcNode.get_name() == 'NI-Controller':
        generateNISources(paramDOM,sourceFolderPath,logTopics)
    else:
        iecommon.logError("The PLC-Type %s is not supported" %plcNode.get_name(), True,logTopics)
    paramDOM.freeDoc()

#=========================================================================
# Entry Point
#=========================================================================
def genPlcSrc(workspacePath, deployName,silecsVersion,logTopics={'errorlog': True}):
    deployFilePath = iefiles.getSilecsDeployFilePath(workspacePath,deployName)
    iecommon.logInfo("Load Deployment document: %s" %deployFilePath,logTopics);

    # Create the source folder if needed
    sourceFolderPath = iefiles.getControllerSourcesDirectory(workspacePath, deployName)
    if not os.path.exists(sourceFolderPath):
        os.makedirs(sourceFolderPath)

    silecsDeploy = Deploy.getDeployFromFile(deployFilePath)
    for controller in silecsDeploy.controllers:
        paramsFile = iefiles.getParameterFile(workspacePath, deployName,controller.hostName)
        generateControllerCode(controller,paramsFile,sourceFolderPath,logTopics)
    
    return "Genplcsrc succeded!"
 
