#!/usr/bin/python
# Copyright 2016 CERN and GSI
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from model.Deploy.Device import *
from model.Deploy.SilecsDesign import *
import libxml2

class Controller(object):

    def __init__(self, xmlNode):
        self.hostName = ""
        self.domain = ""
        self.system = ""
        self.model = ""
        self.protocol = ""
        self.brand = ""
        self.baseAddress = long(0)
        self.xmlNode = None
        self.plcNode = None
        self.devices = []
    
        self.xmlNode = xmlNode
        self.hostName = xmlNode.prop("host-name")
        if xmlNode.hasProp("domain"):
            self.domain = xmlNode.prop("domain")
        self.plcNode = Controller.getPLCNode(xmlNode)
        self.system = self.plcNode.prop('system')
        self.model = self.plcNode.prop('model')
        self.protocol = self.plcNode.prop('protocol')
        
        silecsHeader = Device()
        silecsHeader.silecsDeviceLabel = "SilecsHeader"
        silecsHeader.silecsDesignRef = "SilecsHeader"
        self.devices.append(silecsHeader)
        
        for deviceNode in self.plcNode.xpathEval('Device'):
            device = Device()
            device.initFromXMLNode(deviceNode)
            self.devices.append(device)
    
    def connectDesignObjects(self,silecsDesigns):
        for device in self.devices:
            for design in silecsDesigns:
                if device.silecsDesignRef == design.name:
                    device.silecsDesign = design
    
    def getDevicesOfDesign(self, designName):
        deviceList = []
        for device in self.devices:
            if device.silecsDesignRef == designName:
                deviceList.append(device)
        return deviceList
    
    def getUsedDesigns(self):
        designList = []
        for device in self.devices:
            if not device.silecsDesign in designList:
                designList.append(device.silecsDesign)
        return designList
        
    @staticmethod
    def getPLCNode(controllerNode):
        hostName = controllerNode.prop("host-name")
        plcNodes = controllerNode.xpathEval('*')
        if len(plcNodes) > 1:
            raise Exception( "Error: multiple plc-nodes found for controller: " + hostName )
        if len(plcNodes) < 1:
            raise Exception( "Error: no plc-node found for controller: " + hostName )
        return plcNodes[0]
        
    @staticmethod
    def createController(controllerNode):
        plcNode = Controller.getPLCNode(controllerNode)
        type = plcNode.get_name()
        if   type == "Siemens-PLC":
            return SiemensController(controllerNode)
        elif type == "Schneider-PLC":
            return SchneiderController(controllerNode)
        elif type == "Beckhoff-PLC":
            return BeckhoffController(controllerNode)
        elif type == "Rabbit-uC":
            return RabbitController(controllerNode)
        elif type == "NI-Controller":
            return NIController(controllerNode)
        elif type == "PC-Controller":
            return VirtualS7Controller(controllerNode)
        else:
            raise Exception( "Controller-Type " + type + " not supported" )
    
class SiemensController(Controller):
    def __init__(self, xmlNode):
        super(SiemensController, self).__init__(xmlNode)
        self.brand = "SIEMENS"
        self.baseAddress = long(self.plcNode.prop('base-DB-number'))
        
class BeckhoffController(Controller):
    addressFactor_BC9020 = 2 # BC90xx use 16 bit addressing
    addressFactor_CX9020 = 1
    offset_BC9020 = 32768
    offset_CX9020 = 24576
    def __init__(self, xmlNode):
        super(BeckhoffController, self).__init__(xmlNode)
        self.brand = "BECKHOFF"
        self.baseAddress = long(self.plcNode.prop('base-address'))
        if self.model == 'BC9020':
            self.baseAddress = ( self.baseAddress * self.addressFactor_BC9020 ) + self.offset_BC9020
        elif self.model == 'CX9020':
            self.baseAddress = ( self.baseAddress * self.addressFactor_CX9020 ) + self.offset_CX9020
        else:
            raise Exception( "Unknown beckhoff plc-model:" + self.model )
        
class SchneiderController(Controller):
    addressFactor = 2
    # Premium/Quantum uses 16bit adressing for 32bit values
    # Unity M340 is interpreted as 32bit address
    def __init__(self, xmlNode):
        super(SchneiderController, self).__init__(xmlNode)
        self.brand = "SCHNEIDER"
        self.baseAddress = long(self.plcNode.prop('base-address'))
        self.baseAddress = self.baseAddress * self.addressFactor
        
class RabbitController(Controller):
    addressFactor = 2
    def __init__(self, xmlNode):
        super(RabbitController, self).__init__(xmlNode)
        self.brand = "DIGI"
        self.baseAddress = long(self.plcNode.prop('base-address'))
        self.baseAddress = self.baseAddress * self.addressFactor

class VirtualS7Controller(Controller):
    addressFactor = 1
    def __init__(self, xmlNode):
        super(VirtualS7Controller, self).__init__(xmlNode)
        self.brand = "SIEMENS"
        self.baseAddress = long(self.plcNode.prop('base-address'))
        self.baseAddress = self.baseAddress * self.addressFactor
        
class NIController(Controller):
    def __init__(self, xmlNode):
        super(NIController, self).__init__(xmlNode)
        self.brand = "NI"
        self.baseAddress = 0
        
        
    #xmlNode.shellPrintNode()
